import networkx as nx
import pytest
from networkx.utils import create_py_random_state
from networkx_algo_common_subtree import balanced_embedding
from networkx_algo_common_subtree import maximum_common_ordered_subtree_embedding
from networkx_algo_common_subtree.utils import graph_str
from networkx_algo_common_subtree.utils import random_tree
from networkx_algo_common_subtree._types import OrderedDiGraph
from networkx_algo_common_subtree.tree_embedding import tree_to_seq


def test_null_common_embedding():
    """
    The empty graph is not a tree and should raise an error
    """
    empty = OrderedDiGraph()
    non_empty = random_tree(n=1, create_using=OrderedDiGraph)

    with pytest.raises(nx.NetworkXPointlessConcept):
        maximum_common_ordered_subtree_embedding(empty, empty)

    with pytest.raises(nx.NetworkXPointlessConcept):
        maximum_common_ordered_subtree_embedding(empty, non_empty)

    with pytest.raises(nx.NetworkXPointlessConcept):
        maximum_common_ordered_subtree_embedding(non_empty, empty)


def test_self_common_embedding():
    """
    The common embedding of a tree with itself should always be itself
    """
    rng = create_py_random_state(85652972257)
    for n in range(1, 10):
        tree = random_tree(n=n, seed=rng, create_using=OrderedDiGraph)
        embedding1, embedding2, _ = maximum_common_ordered_subtree_embedding(
            tree, tree
        )
        assert tree.edges == embedding1.edges


def test_common_tree_embedding_small():
    tree1 = OrderedDiGraph([(0, 1)])
    tree2 = OrderedDiGraph([(0, 1), (1, 2)])
    print(graph_str(tree1))
    print(graph_str(tree2))

    embedding1, embedding2, _ = maximum_common_ordered_subtree_embedding(
        tree1, tree2
    )
    print(graph_str(embedding1))
    print(graph_str(embedding2))


def test_common_tree_embedding_small2():
    tree1 = OrderedDiGraph([(0, 1), (2, 3), (4, 5), (5, 6)])
    tree2 = OrderedDiGraph([(0, 1), (1, 2), (0, 3)])
    print(graph_str(tree1))
    print(graph_str(tree2))

    embedding1, embedding2, _ = maximum_common_ordered_subtree_embedding(
        tree1, tree2, node_affinity=None
    )
    print(graph_str(embedding1))
    print(graph_str(embedding2))


def test_all_implementations_are_same():
    """
    Tests several random sequences
    """

    seed = 24658885408229410362279507020239
    rng = create_py_random_state(seed)

    maxsize = 20
    num_trials = 5

    node_affinity = "eq"

    for _ in range(num_trials):
        n1 = rng.randint(1, maxsize)
        n2 = rng.randint(1, maxsize)

        tree1 = random_tree(n1, seed=rng, create_using=OrderedDiGraph)
        tree2 = random_tree(n2, seed=rng, create_using=OrderedDiGraph)

        # Note: the returned sequences may be different (maximum embeddings may
        # not be unique), but the values should all be the same.
        results = {}
        impls = balanced_embedding.available_impls_longest_common_balanced_embedding()
        for impl in impls:
            # FIXME: do we need to rework the return value here?
            embedding1, embedding2, value = maximum_common_ordered_subtree_embedding(
                tree1, tree2, node_affinity=node_affinity, impl=impl
            )
            _check_common_embedding_invariants(tree1, tree2, embedding1, embedding2)
            results[impl] = value

        x = max(results.values())
        assert all(v == x for v in results.values())


def _check_embedding_invariants(tree, subtree):
    assert set(subtree.nodes).issubset(set(tree.nodes)), "must have a node subset"
    assert len(subtree.edges) <= len(tree.edges)


def _check_common_embedding_invariants(tree1, tree2, embedding1, embedding2):
    """
    Validates that this solution satisfies properties of an embedding
    """
    _check_embedding_invariants(tree1, embedding1)
    _check_embedding_invariants(tree2, embedding2)
    assert len(embedding1.nodes) == len(embedding2.nodes)

    _check_contractions(tree1, tree2, embedding1, embedding2)


def _check_contractions(tree1, tree2, embedding1, embedding2):
    """
    Only valid if ``node_affinity`` was "auto"
    """
    # We reverse topologically sort so we always contract from leaves to the
    # root
    remove_nodes1 = [
        n for n in nx.topological_sort(tree1) if n not in embedding1.nodes
    ][::-1]
    remove_nodes2 = [
        n for n in nx.topological_sort(tree2) if n not in embedding2.nodes
    ][::-1]

    contract_edges1 = [(u, v) for v in remove_nodes1 for u in tree1.pred[v]]
    contract_edges2 = [(u, v) for v in remove_nodes2 for u in tree2.pred[v]]

    embedding1_alt = tree1.copy()
    for u, v in contract_edges1:
        nx.contracted_nodes(embedding1_alt, u, v, copy=False, self_loops=False)
    # Handle the case where sources nodes are removed
    embedding1_alt.remove_nodes_from(remove_nodes1)

    embedding2_alt = tree2.copy()
    for u, v in contract_edges2:
        nx.contracted_nodes(embedding2_alt, u, v, copy=False, self_loops=False)
    # Handle the case where sources nodes are removed
    embedding2_alt.remove_nodes_from(remove_nodes2)

    assert set(embedding1.nodes) == set(embedding1_alt.nodes)
    assert set(embedding1.edges) == set(embedding1_alt.edges)

    assert set(embedding2.nodes) == set(embedding2_alt.nodes)
    assert set(embedding2.edges) == set(embedding2_alt.edges)

    if 0:
        graph_str(tree1, write=print)
        graph_str(tree2, write=print)
        graph_str(embedding1, write=print)
        graph_str(embedding2, write=print)
        graph_str(embedding1_alt, write=print)
        graph_str(embedding2_alt, write=print)


def test_forest_case():
    # Test forest case
    F = nx.disjoint_union_all(
        [
            random_tree(3, seed=0, create_using=OrderedDiGraph),
            random_tree(5, seed=1, create_using=OrderedDiGraph),
            random_tree(5, seed=1, create_using=OrderedDiGraph),
            random_tree(2, seed=2, create_using=OrderedDiGraph),
            random_tree(1, seed=3, create_using=OrderedDiGraph),
        ]
    )

    # Just verify everything runs without error for now. Would be best to find
    # something to check though.
    print(graph_str(F))
    sequence, *_ = tree_to_seq(F, item_type="number")
    print("sequence = {!r}".format(sequence))
    sequence, *_ = tree_to_seq(F, item_type="chr")
    print("sequence = {!r}".format(sequence))
    sequence, *_ = tree_to_seq(F, item_type="chr", container_type="str")
    print("sequence = {!r}".format(sequence))
    sequence, *_ = tree_to_seq(F, item_type="chr", container_type="tuple")
    print("sequence = {!r}".format(sequence))



import sys
import networkx as nx


if sys.version_info[0:2] > (3, 6):
    OrderedDiGraph = nx.DiGraph
    OrderedGraph = nx.Graph
else:
    OrderedDiGraph = nx.OrderedDiGraph
    OrderedGraph = nx.OrderedGraph



def test_simple_cases():
    from networkx_algo_common_subtree.balanced_isomorphism import (
        longest_common_balanced_isomorphism,
    )

    open_to_close = {"{": "}", "(": ")", "[": "]"}
    seq1 = "[[]][]"
    seq2 = "[]{}[]"
    res = longest_common_balanced_isomorphism(seq1, seq2, open_to_close)
    (subseq1, subseq2), value = res
    print("subseq1 = {!r}".format(subseq1))
    assert subseq1 == "[][]"

    seq1 = "[[]][]"
    seq2 = "[]{}{}{}{}[]"
    res = longest_common_balanced_isomorphism(seq1, seq2, open_to_close)
    (subseq1, subseq2), value = res
    print("subseq1 = {!r}".format(subseq1))
    assert subseq1 == "[][]"

    seq1 = "[[]]()()()[]"
    seq2 = "[]{}{}{}{}[]"
    res = longest_common_balanced_isomorphism(seq1, seq2, open_to_close)
    (subseq1, subseq2), value = res
    print("subseq1 = {!r}".format(subseq1))
    assert subseq1 == "[][]"

    seq1 = "[[]](){}()()[]"
    seq2 = "[]{}{}{}{}[]"
    res = longest_common_balanced_isomorphism(seq1, seq2, open_to_close)
    (subseq1, subseq2), value = res
    print("subseq1 = {!r}".format(subseq1))
    assert subseq1 == "[]{}[]"

    seq1 = "()([])[]"
    seq2 = "[](){}()"
    res = longest_common_balanced_isomorphism(seq1, seq2, open_to_close)
    (subseq1, subseq2), value = res
    print("subseq1 = {!r}".format(subseq1))
    assert subseq1 == "()()"

    seq1 = "[()([])[]]"
    seq2 = "{[](){}()}"
    res = longest_common_balanced_isomorphism(seq1, seq2, open_to_close)
    (subseq1, subseq2), value = res
    print("subseq1 = {!r}".format(subseq1))
    assert subseq1 == "()()"


def test_all_implementations_are_same():
    """
    Tests several random sequences
    """
    from networkx_algo_common_subtree import balanced_isomorphism
    from networkx_algo_common_subtree.balanced_sequence import random_balanced_sequence
    from networkx.utils import create_py_random_state

    seed = 93024896892223032652928827097264
    rng = create_py_random_state(seed)

    maxsize = 20
    num_trials = 5

    for _ in range(num_trials):
        n1 = rng.randint(1, maxsize)
        n2 = rng.randint(1, maxsize)

        seq1, open_to_close = random_balanced_sequence(n1, seed=rng)
        seq2, open_to_close = random_balanced_sequence(
            n2, open_to_close=open_to_close, seed=rng
        )
        longest_common_balanced_sequence = (
            balanced_isomorphism.longest_common_balanced_isomorphism
        )

        # Note: the returned sequences may be different (maximum embeddings may not
        # be unique), but the values should all be the same.
        results = {}
        impls = (
            balanced_isomorphism.available_impls_longest_common_balanced_isomorphism()
        )
        for impl in impls:
            best, val = longest_common_balanced_sequence(
                seq1, seq2, open_to_close, node_affinity=None, impl=impl
            )
            results[impl] = val
        first = next(iter(results.values()))
        if not all(v == first for v in results.values()):
            print("results = {!r}".format(results))
            raise AssertionError("Implementations returned different values")



"""
Algorithm extensions
"""
__version__ = '0.2.2'

__mkinit__ = """
mkinit -m networkx_algo_common_subtree -w
"""

__submodules__ = {
    'utils': [],
    'balanced_embedding': [],
    'balanced_isomorphism': [],
    'tree_embedding': ['maximum_common_ordered_subtree_embedding'],
    'tree_isomorphism': ['maximum_common_ordered_subtree_isomorphism'],
}

from networkx_algo_common_subtree import balanced_embedding
from networkx_algo_common_subtree import balanced_isomorphism
from networkx_algo_common_subtree import tree_embedding
from networkx_algo_common_subtree import tree_isomorphism
from networkx_algo_common_subtree import utils

from networkx_algo_common_subtree.tree_embedding import (
    maximum_common_ordered_subtree_embedding,)
from networkx_algo_common_subtree.tree_isomorphism import (
    maximum_common_ordered_subtree_isomorphism,)

__all__ = ['balanced_embedding', 'balanced_isomorphism',
           'maximum_common_ordered_subtree_embedding',
           'maximum_common_ordered_subtree_isomorphism', 'tree_embedding',
           'tree_isomorphism', 'utils']



import operator
from .balanced_sequence import (
    generate_balance_unsafe,
    IdentityDict,
)


def available_impls_longest_common_balanced_isomorphism():
    """
    Returns all available implementations for
    :func:`longest_common_balanced_isomorphism`.

    Returns
    -------
    List[str]
        the string code for each available implementation
    """
    impls = ["recurse", "iter"]
    if _cython_lcsi_backend():
        impls += [
            "iter-cython",
        ]
    return impls


def longest_common_balanced_isomorphism(
    seq1, seq2, open_to_close, open_to_node=None, node_affinity="auto", impl="auto"
):
    r"""
    Finds the longest common balanced sequence isomorphism between two
    sequences.

    Parameters
    ----------
    seq1, seq2: Sequence[TokT]
        two input balanced sequences

    open_to_close : Dict[TokT, TokT]
        a mapping from opening to closing tokens in the balanced sequence

    open_to_node : Dict[TokT, Any] | None
        If unspecified an identity mapping is assumed. Otherwise this is a
        dictionary that maps a sequence token to a value which is used in the
        ``node_affinity`` comparison.  Typically these are values corresponding
        to an original problem (e.g. a tree node).  This should only used in
        the case where the tokens in each sequence ``seq1`` and ``seq2`` are
        unique. NOTE: in the case where sequence tokens are not unique,
        sequences can always be re-encoded to differentiate between the same
        token at different indices without loss of generality.

    node_affinity : None | str | callable
        Function for to determine if two nodes can be matched. The function
        should take two arguments `node1` and `node2` and return a non-negative
        affinity score that is zero if the nodes are not allowed to match and
        some positive value indicating the strength of the match.  The return
        is interpreted as a weight that is used to break ties. If
        ``node_affinity=None`` then any node can match any other node and only
        the topology is important.  The default is "auto", which will use
        ``operator.eq`` to do a simple equality test on the nodes.

    impl : str
        Determines the backend implementation. Available choices are given by
        :func:`available_impls_longest_common_balanced_isomorphism`.
        The default is "auto", which chooses "iter-cython" if available,
        otherwise "iter".

    Returns
    -------
    Tuple[Tuple[Sequence[TokT], Sequence[TokT]], Float]
        A tuple indicating the common subsequence isomorphism of sequence1 and
        sequence2 (usually these are the same) and its value.

    See Also
    --------
    * This function is used to implement :func:`tree_isomorphism.maximum_common_ordered_subtree_isomorphism`
    * A similar function that relaxes isomorphisms to embeddings is :func:`networkx.algorithms.strength.balanced_sequence.longest_common_balanced_sequence`

    Example
    -------
    >>> # Given two sequences and a mapping between opening and closing tokens
    >>> # we find the longest common subsequence (achievable by repeated
    >>> # balanced decomposition)
    >>> seq1 = "[][[]][]"
    >>> seq2 = "[[]][[]]"
    >>> open_to_close = {"[": "]"}
    >>> best, value = longest_common_balanced_isomorphism(seq1, seq2, open_to_close)
    ...
    >>> subseq1, subseq2 = best
    >>> print("subseq1 = {!r}".format(subseq1))
    subseq1 = '[[]][]'

    >>> # 1-label case from the paper (see Example 5)
    >>> # https://pdfs.semanticscholar.org/0b6e/061af02353f7d9b887f9a378be70be64d165.pdf
    >>> seq1 = "0010010010111100001011011011"
    >>> seq2 = "001000101101110001000100101110111011"
    >>> open_to_close = {"0": "1"}
    >>> best, value = longest_common_balanced_isomorphism(seq1, seq2, open_to_close)
    >>> subseq1, subseq2 = best
    >>> print("subseq1 = {!r}".format(subseq1))
    subseq1 = '001000101111000010111011'
    >>> assert value == 12

    >>> # 3-label case
    >>> seq1 = "{({})([[]([]){(()(({()[]({}{})}))){}}])}"
    >>> seq2 = "{[({{}}{{[][{}]}(()[(({()}[])){[]()}])})]}"
    >>> open_to_close = {"{": "}", "(": ")", "[": "]"}
    >>> best, value = longest_common_balanced_isomorphism(seq1, seq2, open_to_close)
    >>> subseq1, subseq2 = best
    >>> print("subseq1 = {!r}".format(subseq1))
    subseq1 = '{([(){()}])}'
    >>> assert value == 6
    """
    if node_affinity == "auto" or node_affinity == "eq":
        node_affinity = operator.eq
    if node_affinity is None:

        def _matchany(a, b):
            return True

        node_affinity = _matchany
    if open_to_node is None:
        open_to_node = IdentityDict()
    full_seq1 = seq1
    full_seq2 = seq2
    if impl == "auto":
        if _cython_lcsi_backend(error="ignore"):
            impl = "iter-cython"
        else:
            impl = "iter"

    if impl == "iter":
        val_any, best_any, val_lvl, best_lvl = _lcsi_iter(
            full_seq1,
            full_seq2,
            open_to_close,
            node_affinity,
            open_to_node,
        )
    elif impl == "iter-cython":
        balanced_isomorphism_cython = _cython_lcsi_backend(error="raise")
        (
            val_any,
            best_any,
            val_lvl,
            best_lvl,
        ) = balanced_isomorphism_cython._lcsi_iter_cython(
            full_seq1, full_seq2, open_to_close, node_affinity, open_to_node
        )
    elif impl == "recurse":
        _memo = {}
        _seq_memo = {}
        val_any, best_any, val_lvl, best_lvl = _lcsi_recurse(
            full_seq1,
            full_seq2,
            open_to_close,
            node_affinity,
            open_to_node,
            _memo,
            _seq_memo,
        )
    else:
        raise KeyError(impl)

    best = best_any
    value = val_any

    return best, value


def _lcsi_iter(full_seq1, full_seq2, open_to_close, node_affinity, open_to_node):
    """
    Converts :func:`_lcsi_recurse` into an iterative algorithm.

    Example
    -------
    >>> import operator as op
    >>> seq1 = full_seq1 = '[[]][]'
    >>> seq2 = full_seq2 = '[]{}[]'
    >>> open_to_close = {"{": "}", "(": ")", "[": "]"}
    >>> node_affinity = op.eq
    >>> _memo, _seq_memo = {}, {}
    >>> open_to_node = IdentityDict()
    >>> res = _lcsi_iter(full_seq1, full_seq2, open_to_close, node_affinity,
    ...                 open_to_node)
    >>> value, best, *_ = _lcsi_iter(
    ...     full_seq1, full_seq2, open_to_close, node_affinity, open_to_node)
    >>> print('value = {!r}, best = {!r}'.format(value, best[0]))
    value = 2, best = '[][]'
    """
    all_decomp1 = generate_all_decomp_nocat(full_seq1, open_to_close, open_to_node)
    all_decomp2 = generate_all_decomp_nocat(full_seq2, open_to_close, open_to_node)
    key0 = (full_seq1, full_seq2)
    frame0 = key0
    stack = [frame0]

    # Memoize mapping (seq1, seq2) -> best size, embeddings
    _results = {}

    # Populate base cases
    empty1 = type(next(iter(all_decomp1.keys())))()
    empty2 = type(next(iter(all_decomp2.keys())))()
    best = (empty1, empty2)
    base_result = (0, best, 0, best)
    for seq1 in all_decomp1.keys():
        key1 = seq1
        t1, a1, b1, head1, tail1 = all_decomp1[key1]
        _results[(seq1, empty2)] = base_result
        _results[(head1, empty2)] = base_result
        _results[(tail1, empty2)] = base_result

    for seq2 in all_decomp2.keys():
        key2 = seq2
        t2, a2, b2, head2, tail2 = all_decomp2[key2]
        _results[(empty1, seq2)] = base_result
        _results[(empty1, head2)] = base_result
        _results[(empty1, tail2)] = base_result

    while stack:
        key = stack[-1]
        if key not in _results:
            seq1, seq2 = key

            t1, a1, b1, head1, tail1 = all_decomp1[seq1]
            t2, a2, b2, head2, tail2 = all_decomp2[seq2]

            best_any = None
            best_lvl = None
            val_any = 0
            val_lvl = 0

            # When using the head part of the decomp, we can only update the "low" candidate
            try_key = (head1, seq2)
            if try_key in _results:
                val_any_h1s2, cand_any_h1s2, _, _ = _results[try_key]
            else:
                stack.append(try_key)
                continue

            try_key = (tail1, seq2)
            if try_key in _results:
                val_any_t1s2, cand_any_t1s2, val_lvl_t1s2, cand_lvl_t1s2 = _results[
                    try_key
                ]
            else:
                stack.append(try_key)
                continue

            try_key = (seq1, head2)
            if try_key in _results:
                val_any_s1h2, cand_any_s1h2, _, _ = _results[try_key]
            else:
                stack.append(try_key)
                continue

            try_key = (seq1, tail2)
            if try_key in _results:
                val_any_s1t2, cand_any_s1t2, val_lvl_s1t2, cand_lvl_s1t2 = _results[
                    try_key
                ]
            else:
                stack.append(try_key)
                continue

            if val_any_h1s2 > val_any:
                val_any = val_any_h1s2
                best_any = cand_any_h1s2

            if val_any_t1s2 > val_any:
                val_any = val_any_t1s2
                best_any = cand_any_t1s2

            if val_any_s1h2 > val_any:
                val_any = val_any_s1h2
                best_any = cand_any_s1h2

            if val_any_s1t2 > val_any:
                val_any = val_any_s1t2
                best_any = cand_any_s1t2

            # The "LVL" case should include the case where any match exists on this
            # level. That means as long as we don't consider the heads, tail
            # matches are fine.
            if val_lvl_s1t2 > val_lvl:
                val_lvl = val_lvl_s1t2
                best_lvl = cand_lvl_s1t2

            if val_lvl_t1s2 > val_lvl:
                val_lvl = val_lvl_t1s2
                best_lvl = cand_lvl_t1s2

            # Case 1: The LCS involves this edge
            affinity = node_affinity(t1, t2)
            if affinity:
                try_key = (head1, head2)
                if try_key in _results:
                    _, _, pval_lvl_h1h2, new_lvl_h1h2 = _results[try_key]
                else:
                    stack.append(try_key)
                    continue

                try_key = (tail1, tail2)
                if try_key in _results:
                    _, _, pval_lvl_t1t2, new_lvl_t1t2 = _results[try_key]
                else:
                    stack.append(try_key)
                    continue

                # Add to the best solution at the former level
                new_val_lvl = pval_lvl_h1h2 + pval_lvl_t1t2 + affinity
                if new_val_lvl > val_lvl:
                    val_lvl = new_val_lvl
                    new_head1, new_head2 = new_lvl_h1h2
                    new_tail1, new_tail2 = new_lvl_t1t2
                    subseq1 = a1 + new_head1 + b1 + new_tail1
                    subseq2 = a2 + new_head2 + b2 + new_tail2
                    best_lvl = (subseq1, subseq2)

            # If the current level is better than any of the nestings forget the
            # nestings (we can never improve them) and just use the level
            if val_lvl >= val_any:
                val_any = val_lvl
                best_any = best_lvl

            if best_lvl is None:
                best_lvl = (empty1, empty2)

            if best_any is None:
                best_any = (empty1, empty2)

            # We solved the frame
            found = (val_any, best_any, val_lvl, best_lvl)
            _results[key] = found
        stack.pop()

    found = _results[key0]
    return found


def _lcsi_recurse(
    seq1, seq2, open_to_close, node_affinity, open_to_node, _memo, _seq_memo
):
    """
    Recursive implementation of longest common substring isomorphism.

    Notes
    -----

    Recall a balanced sequence ``s`` can be decomposed as follows:

    .. code::
        # Input balanced sequence s
        s      = '([()[]])[{}([[]])]'

        # Its decomposition into a head and tail
        s.a    = '('
        s.head =  '[()[]]'
        s.b    =        ')'
        s.tail =         '[{}([[]])]'


        # A recursive tail decomposition of a sequence


    The recurrence returns two values: (0) the best isomorphism that includes
    the start of one of the input sequences and (1) the best isomorphism at a
    deeper location in the sequence.  Is defined as follows:

    .. code::

        Let lcsi(s1, s2, LVL) be the best isoseq between s1 and s2 that includes
            either the token at s1[0] and s2[0] or some token in the recursive
            tail decomposition of each sequence. (i.e. the lsci begins at some
            node at the current outermost level of nesting)

        Let lcsi(s1, s2, ANY) be the best overall isoseq between s1 and s2 that
            might exist in the head of one of the sequences in the recursive
            tail decompositions of s1 and s2. (i.e. the lsci might begin at
            some deeper nesting level in either sequence).

        lcsi(s1, '', ANY) = 0
        lcsi(s1, '', LVL) = 0
        lcsi('', s2, ANY) = 0
        lcsi('', s2, LVL) = 0

        # The value of the LCSI including the a first token match is 0 if the
        # tokens dont match otherwise it is the affinity plus the LCSI that
        # includes the next token in both the head and tail of the balanced
        # sequence.
        #
        # IT CAN ALSO be the case that one string matches the tail of another
        lcsi(s1, s2, LVL) = max(
            lcsi(s1, s2.tail, LVL) + lcsi(s1.tail, s2, LVL)
            lcsi(s1.head, s2.head, LVL) + lcsi(s1.tail, s2.tail, LVL) + affinity(s1.a, s2.a) if affinity(s1.a, s2.a) else 0
        )
        # Note that we cannot consider any exclusion cases because we are not
        # allowed to "skip" edges like we are in the "subsequence embedding"
        # problem.

        # For the LCSI that excludes the current matching token, we peel that
        # token off of the first and second sequence and subproblems that
        # compare the head or tail of one sequence to the entire other
        # sequence. Because the current leading token is discarded in at least
        # one of the input sequences we consider the include and exclude case
        # for all subproblems here.
        lcsi(s1, s2, ANY) = max(
            #
            lcsi(s1, s2, LVL)

            # The case where we only consider the head/tail of s1
            lcsi(s1.head, s2, ANY),
            lcsi(s1.tail, s2, ANY),

            # The case where we only consider the head/tail of s1
            lcsi(s1.head, s2, ANY),
            lcsi(s1.tail, s2, ANY),
        )

        # Note that by the way the recurrence is defined, s1.head will be
        # compared to s2.head in subsequent subproblems, so explicitly adding
        # that decomposition here is not necessary.

        The final lcsi for s1 and s2 is

        lcsi(s1, s2) = lcsi(s1, s2, ANY)

    Example
    -------
    >>> import operator as op
    >>> node_affinity = op.eq
    >>> open_to_close = {"{": "}", "(": ")", "[": "]"}
    >>> open_to_node = IdentityDict()
    >>> # ---
    >>> seq1 = full_seq1 = "[][[]][]"
    >>> seq2 = full_seq2 = "[[]][[]]"
    >>> _memo, _seq_memo = {}, {}
    >>> value, best, *_ = _lcsi_recurse(
    ...     full_seq1, full_seq2, open_to_close, node_affinity, open_to_node,
    ...     _memo, _seq_memo)
    >>> print('value = {!r}, best = {!r}'.format(value, best[0]))
    value = 3, best = '[[]][]'
    >>> # ---
    >>> seq1 = full_seq1 = "[{[[]]}]"
    >>> seq2 = full_seq2 = "[[{[[]]}]]"
    >>> _memo, _seq_memo = {}, {}
    >>> value, best, *_ = _lcsi_recurse(
    ...     full_seq1, full_seq2, open_to_close, node_affinity, open_to_node,
    ...     _memo, _seq_memo)
    >>> print('value = {!r}, best = {!r}'.format(value, best[0]))
    value = 4, best = '[{[[]]}]'
    >>> # ---
    >>> seq1 = full_seq1 = '({{{[]}}})'
    >>> seq2 = full_seq2 = '[{{([()])}}]'
    >>> _memo, _seq_memo = {}, {}
    >>> value, best, *_ = _lcsi_recurse(
    ...     full_seq1, full_seq2, open_to_close, node_affinity, open_to_node,
    ...     _memo, _seq_memo)
    >>> print('value = {!r}, best = {!r}'.format(value, best[0]))
    value = 2, best = '{{}}'
    >>> # ---
    >>> full_seq1 = '[[]][]'
    >>> full_seq2 = '[]{}[]'
    >>> _memo, _seq_memo = {}, {}
    >>> value, best, *_ = _lcsi_recurse(
    ...     full_seq1, full_seq2, open_to_close, node_affinity, open_to_node,
    ...     _memo, _seq_memo)
    >>> print('value = {!r}, best = {!r}'.format(value, best[0]))
    value = 2, best = '[][]'
    >>> # ---
    >>> full_seq1 = '[[]][]'
    >>> full_seq2 = '[]{}{}{}{}[]'
    >>> _memo, _seq_memo = {}, {}
    >>> value, best, *_ = _lcsi_recurse(
    ...     full_seq1, full_seq2, open_to_close, node_affinity, open_to_node,
    ...     _memo, _seq_memo)
    >>> print('value = {!r}, best = {!r}'.format(value, best[0]))
    value = 2, best = '[][]'
    """
    if not seq1:
        return 0, (seq1, seq1), 0, (seq1, seq1)
    elif not seq2:
        return 0, (seq2, seq2), 0, (seq2, seq2)
    else:
        key1 = hash(seq1)
        key2 = hash(seq2)
        key = hash((key1, key2))
        if key in _memo:
            return _memo[key]

        if key1 in _seq_memo:
            a1, b1, head1, tail1 = _seq_memo[key1]
        else:
            a1, b1, head1, tail1 = balanced_decomp_unsafe_nocat(seq1, open_to_close)
            _seq_memo[key1] = a1, b1, head1, tail1

        if key2 in _seq_memo:
            a2, b2, head2, tail2 = _seq_memo[key2]
        else:
            a2, b2, head2, tail2 = balanced_decomp_unsafe_nocat(seq2, open_to_close)
            _seq_memo[key2] = a2, b2, head2, tail2

        # TODO: IS THIS THE CORRECT MODIFICATION TO THE RECURRANCE TO
        # ACHIEVE A SUBTREE ISOMORPHISM INSTEAD OF AN EMBEDDING?
        r"""
        We return two solutions at each step: the solution value at
        this level if one exists, and the solution value at any other depth.
        We are allowed to add to the first, but can take the second if we want
        to.

        This should work because we know a solution that skipped a layer will
        never be added to, and we are always keeping track of the solution that
        might change. By the time we get to the root level, we have enough info
        to know which is better.
        """

        # Consider the case where the best isoseq does not include the leading
        # tokens of s1 and s2.
        best_any = None
        best_lvl = None
        val_any = 0
        val_lvl = 0

        # When using the head part of the decomp, we can only update the "low" candidate
        val_any_h1s2, cand_any_h1s2, _, _, = _lcsi_recurse(
            head1, seq2, open_to_close, node_affinity, open_to_node, _memo, _seq_memo
        )
        val_any_t1s2, cand_any_t1s2, val_lvl_t1s2, cand_lvl_t1s2 = _lcsi_recurse(
            tail1, seq2, open_to_close, node_affinity, open_to_node, _memo, _seq_memo
        )
        val_any_s1h2, cand_any_s1h2, _, _ = _lcsi_recurse(
            seq1, head2, open_to_close, node_affinity, open_to_node, _memo, _seq_memo
        )
        val_any_s1t2, cand_any_s1t2, val_lvl_s1t2, cand_lvl_s1t2 = _lcsi_recurse(
            seq1, tail2, open_to_close, node_affinity, open_to_node, _memo, _seq_memo
        )

        if val_any_h1s2 > val_any:
            val_any = val_any_h1s2
            best_any = cand_any_h1s2

        if val_any_t1s2 > val_any:
            val_any = val_any_t1s2
            best_any = cand_any_t1s2

        if val_any_s1h2 > val_any:
            val_any = val_any_s1h2
            best_any = cand_any_s1h2

        if val_any_s1t2 > val_any:
            val_any = val_any_s1t2
            best_any = cand_any_s1t2

        # The "LVL" case should include the case where any match exists on this
        # level. That means as long as we don't consider the heads, tail
        # matches are fine.
        if val_lvl_s1t2 > val_lvl:
            val_lvl = val_lvl_s1t2
            best_lvl = cand_lvl_s1t2

        if val_lvl_t1s2 > val_lvl:
            val_lvl = val_lvl_t1s2
            best_lvl = cand_lvl_t1s2

        # Consider the case where the best isoseq does include the leading
        # tokens of s1 and s2.
        t1 = open_to_node[a1[0]]
        t2 = open_to_node[a2[0]]
        affinity = node_affinity(t1, t2)
        if affinity:

            # Note, the "ex" portions of the LCSI don't matter because val_any
            # and best_any will already contain them if they matter We only care
            # about extending the current "in" case.
            # (Actually this is wrong)
            _, _, pval_lvl_h1h2, new_lvl_h1h2 = _lcsi_recurse(
                head1,
                head2,
                open_to_close,
                node_affinity,
                open_to_node,
                _memo,
                _seq_memo,
            )

            _, _, pval_lvl_t1t2, new_lvl_t1t2 = _lcsi_recurse(
                tail1,
                tail2,
                open_to_close,
                node_affinity,
                open_to_node,
                _memo,
                _seq_memo,
            )

            # Add to the best solution at the former level
            new_val_lvl = pval_lvl_h1h2 + pval_lvl_t1t2 + affinity
            if new_val_lvl > val_lvl:
                val_lvl = new_val_lvl
                new_head1, new_head2 = new_lvl_h1h2
                new_tail1, new_tail2 = new_lvl_t1t2
                subseq1 = a1 + new_head1 + b1 + new_tail1
                subseq2 = a2 + new_head2 + b2 + new_tail2
                best_lvl = (subseq1, subseq2)

        # If the current level is better than any of the nestings forget the
        # nestings (we can never improve them) and just use the level
        if val_lvl >= val_any:
            val_any = val_lvl
            best_any = best_lvl

        if best_lvl is None:
            best_lvl = (type(seq1)(), type(seq2)())

        if best_any is None:
            best_any = (type(seq1)(), type(seq2)())

        # We return two solutions:
        # the best that includes any token at this level (lvl)
        # the best overall that could be at a deeper nesting (nst)
        found = (val_any, best_any, val_lvl, best_lvl)
        _memo[key] = found
        return found


def balanced_decomp_unsafe_nocat(sequence, open_to_close):
    """
    Similar to :func:`balanced_decomp` but assumes that ``sequence`` is valid
    balanced sequence in order to execute faster. Also does not return
    the concatenated head_tail as it is unused in the isomorphim problem.

    SeeAlso
    -------
    balanced_decomp, balanced_decomp_unsafe
    """
    gen = generate_balance_unsafe(sequence, open_to_close)

    bal_curr, tok_curr = next(gen)
    pop_open = sequence[0:1]
    want_close = open_to_close[tok_curr]

    head_stop = 1
    for head_stop, (bal_curr, tok_curr) in enumerate(gen, start=1):
        if bal_curr and tok_curr == want_close:
            pop_close = sequence[head_stop : head_stop + 1]
            break
    head = sequence[1:head_stop]
    tail = sequence[head_stop + 1 :]
    return pop_open, pop_close, head, tail


def generate_all_decomp_nocat(seq, open_to_close, open_to_node=None):
    """
    Generates all decompositions of a single balanced sequence by recursive
    decomposition of the head, tail.

    Parameters
    ----------
    seq : Tuple | str
        a tuple of hashable items or a string where each character is an item

    open_to_close : Dict
        a dictionary that maps opening tokens to closing tokens in the balanced
        sequence problem.

    open_to_node : Dict
        a dictionary that maps a sequence token to a token corresponding to an
        original problem (e.g. a tree node)

    Returns
    -------
    Dict :
        mapping from a sub-sequence to its decomposition

    SeeAlso
    -------
    generate_balance_unsafe, generate_balance

    Example
    -------
    >>> import pprint
    >>> seq = '{{(){}}}'
    >>> open_to_close = {'[': ']', '{': '}', '(': ')'}
    >>> all_decomp = generate_all_decomp_nocat(seq, open_to_close)
    >>> print('all_decomp = {}'.format(pprint.pformat(all_decomp)))

    all_decomp = {'(){}': ('(', '(', ')', '', '{}'),
     '{(){}}': ('{', '{', '}', '(){}', ''),
     '{{(){}}}': ('{', '{', '}', '{(){}}', ''),
     '{}': ('{', '{', '}', '', '')}
    """
    if open_to_node is None:
        open_to_node = IdentityDict()
    all_decomp = {}
    stack = [seq]
    while stack:
        seq = stack.pop()
        if seq not in all_decomp and seq:
            (pop_open, pop_close, head, tail) = balanced_decomp_unsafe_nocat(
                seq, open_to_close
            )
            node = open_to_node[pop_open[0]]
            all_decomp[seq] = (node, pop_open, pop_close, head, tail)
            if head:
                if tail:
                    stack.append(tail)
                stack.append(head)
            elif tail:
                stack.append(tail)
    return all_decomp


def _cython_lcsi_backend(error="ignore", verbose=0):
    """
    Returns the cython backend if available, otherwise None

    CommandLine
    -----------
    xdoctest -m networkx_algo_common_subtree.balanced_isomorphism _cython_lcsi_backend
    """
    try:
        # Attempt to use the module build with CMake
        from . import balanced_isomorphism_cython
    except Exception:
        if error == "ignore":
            balanced_isomorphism_cython = None
        elif error == "raise":
            raise
        else:
            raise KeyError(error)
    return balanced_isomorphism_cython



"""
Helpers and utilities for balanced sequence problems.
"""

__all__ = [
    "random_balanced_sequence",
]


class UnbalancedException(Exception):
    """
    Denotes that a sequence was unbalanced
    """

    pass


class IdentityDict:
    """
    Used when ``open_to_node`` is unspecified
    """

    def __getitem__(self, key):
        return key


def generate_all_decomp(seq, open_to_close, open_to_node=None):
    """
    Generates all decompositions of a single balanced sequence by
    recursive decomposition of the head, tail, and head|tail.

    Parameters
    ----------
    seq : Tuple | str
        a tuple of hashable items or a string where each character is an item

    open_to_close : Dict
        a dictionary that maps opening tokens to closing tokens in the balanced
        sequence problem.

    open_to_node : Dict
        a dictionary that maps a sequence token to a token corresponding to an
        original problem (e.g. a tree node)

    Returns
    -------
    Dict :
        mapping from a sub-sequence to its decomposition

    Notes
    -----
    In the paper: See Definition 2, 4, Lemma, 1, 2, 3, 4.

    Example
    -------
    >>> # Example 2 in the paper (one from each column)
    >>> seq = "00100100101111"
    >>> open_to_close = {"0": "1"}
    >>> all_decomp = generate_all_decomp(seq, open_to_close)
    >>> assert len(all_decomp) == len(seq) // 2
    >>> import pprint
    >>> pprint.pprint(all_decomp)
    {'00100100101111': ('0', '0', '1', '010010010111', '', '010010010111'),
     '0010010111': ('0', '0', '1', '01001011', '', '01001011'),
     '001011': ('0', '0', '1', '0101', '', '0101'),
     '01': ('0', '0', '1', '', '', ''),
     '010010010111': ('0', '0', '1', '', '0010010111', '0010010111'),
     '01001011': ('0', '0', '1', '', '001011', '001011'),
     '0101': ('0', '0', '1', '', '01', '01')}

    Example
    -------
    >>> open_to_close = {"{": "}", "(": ")", "[": "]"}
    >>> seq = "({[[]]})[[][]]{{}}"
    >>> all_decomp = generate_all_decomp(seq, open_to_close)
    >>> node, *decomp = all_decomp[seq]
    >>> pop_open, pop_close, head, tail, head_tail = decomp
    >>> print("node = {!r}".format(node))
    node = '('
    >>> print("pop_open = {!r}".format(pop_open))
    pop_open = '('
    >>> print("pop_close = {!r}".format(pop_close))
    pop_close = ')'
    >>> print("head = {!r}".format(head))
    head = '{[[]]}'
    >>> print("tail = {!r}".format(tail))
    tail = '[[][]]{{}}'
    >>> print("head_tail = {!r}".format(head_tail))
    head_tail = '{[[]]}[[][]]{{}}'
    >>> decomp_alt = balanced_decomp(seq, open_to_close)
    >>> assert decomp_alt == tuple(decomp)

    Example
    -------
    >>> seq, open_to_close = random_balanced_sequence(10)
    >>> all_decomp = generate_all_decomp(seq, open_to_close)
    """
    if open_to_node is None:
        open_to_node = IdentityDict()
    all_decomp = {}
    stack = [seq]
    while stack:
        seq = stack.pop()
        if seq not in all_decomp and seq:
            (pop_open, pop_close, head, tail, head_tail) = balanced_decomp_unsafe(
                seq, open_to_close
            )
            node = open_to_node[pop_open[0]]
            all_decomp[seq] = (node, pop_open, pop_close, head, tail, head_tail)
            if head:
                if tail:
                    stack.append(head_tail)
                    stack.append(tail)
                stack.append(head)
            elif tail:
                stack.append(tail)
    return all_decomp


def balanced_decomp(sequence, open_to_close):
    """
    Generates a decomposition of a balanced sequence.

    Parameters
    ----------
    sequence : str | Tuple
        balanced sequence to be decomposed

    open_to_close: dict
        a dictionary that maps opening tokens to closing tokens in the balanced
             sequence problem.

    Returns
    -------
    : tuple[SeqT, SeqT, SeqT, SeqT, SeqT]
        where ``SeqT = type(sequence)``
        Contents of this tuple are:

            0. a1 - a sequence of len(1) containing the current opening token
            1. b1 - a sequence of len(1) containing the current closing token
            2. head - head of the sequence
            3. tail - tail of the sequence
            4. head_tail - the concatenated head and tail

    Example
    -------
    >>> # Example 3 from the paper
    >>> sequence = "001000101101110001000100101110111011"
    >>> open_to_close = {"0": "1"}
    >>> a1, b1, head, tail, head_tail = balanced_decomp(sequence, open_to_close)
    >>> print("head = {!r}".format(head))
    head = '010001011011'
    >>> print("tail = {!r}".format(tail))
    tail = '0001000100101110111011'

    Example
    -------
    >>> open_to_close = {0: 1}
    >>> sequence = [0, 0, 0, 1, 1, 1, 0, 1]
    >>> a1, b1, head, tail, head_tail = balanced_decomp(sequence, open_to_close)
    >>> print("a1 = {!r}".format(a1))
    a1 = [0]
    >>> print("b1 = {!r}".format(b1))
    b1 = [1]
    >>> print("head = {!r}".format(head))
    head = [0, 0, 1, 1]
    >>> print("tail = {!r}".format(tail))
    tail = [0, 1]
    >>> print("head_tail = {!r}".format(head_tail))
    head_tail = [0, 0, 1, 1, 0, 1]
    >>> a2, b2, tail1, tail2, head_tail2 = balanced_decomp(tail, open_to_close)

    Example
    -------
    >>> open_to_close = {"{": "}", "(": ")", "[": "]"}
    >>> sequence = "({[[]]})[[][]]"
    >>> a1, b1, head, tail, head_tail = balanced_decomp(sequence, open_to_close)
    >>> print("a1 = {!r}".format(a1))
    a1 = '('
    >>> print("b1 = {!r}".format(b1))
    b1 = ')'
    >>> print("head = {!r}".format(head))
    head = '{[[]]}'
    >>> print("tail = {!r}".format(tail))
    tail = '[[][]]'
    >>> print("head_tail = {!r}".format(head_tail))
    head_tail = '{[[]]}[[][]]'
    >>> a2, b2, tail1, tail2, head_tail2 = balanced_decomp(tail, open_to_close)
    >>> print("a2 = {!r}".format(a2))
    a2 = '['
    >>> print("b2 = {!r}".format(b2))
    b2 = ']'
    >>> print("tail1 = {!r}".format(tail1))
    tail1 = '[][]'
    >>> print("tail2 = {!r}".format(tail2))
    tail2 = ''
    >>> print("head_tail2 = {!r}".format(head_tail2))
    head_tail2 = '[][]'
    """
    gen = generate_balance(sequence, open_to_close)

    bal_curr, tok_curr, _ = next(gen)
    pop_open = sequence[0:1]
    want_close = open_to_close[tok_curr]

    head_stop = 1
    for head_stop, (bal_curr, tok_curr, _) in enumerate(gen, start=1):
        if tok_curr is None:
            break
        elif bal_curr and tok_curr == want_close:
            pop_close = sequence[head_stop : head_stop + 1]
            break
    head = sequence[1:head_stop]
    tail = sequence[head_stop + 1 :]
    head_tail = head + tail
    return pop_open, pop_close, head, tail, head_tail


def generate_balance(sequence, open_to_close):
    r"""
    Iterates through a balanced sequence and reports if the sequence-so-far
    is balanced at that position or not.

    Parameters
    ----------
    sequence: Iterable[TokT]:
        an input balanced sequence

    open_to_close : Dict[TokT, TokT]
        a mapping from opening to closing tokens in the balanced sequence

    Yields
    ------
    Tuple[bool, TokT, int]:
        boolean indicating if the sequence is balanced at this index, and the
        current token, and the index of the matching opening token.

    Raises
    ------
    UnbalancedException - if the input sequence is not balanced

    Example
    -------
    >>> open_to_close = {0: 1}
    >>> sequence = [0, 0, 0, 1, 1, 1]
    >>> gen = list(generate_balance(sequence, open_to_close))
    >>> for flag, token, idx in gen:
    ...     print("flag={:d}, token={}, prev_idx={}".format(flag, token, idx))
    flag=0, token=0, prev_idx=None
    flag=0, token=0, prev_idx=None
    flag=0, token=0, prev_idx=None
    flag=0, token=1, prev_idx=2
    flag=0, token=1, prev_idx=1
    flag=1, token=1, prev_idx=0

    Example
    -------
    >>> sequence, open_to_close = random_balanced_sequence(4, seed=0)
    >>> gen = list(generate_balance(sequence, open_to_close))
    """
    stack = []
    # Traversing the Expression
    for idx, token in enumerate(sequence):

        if token in open_to_close:
            # Push opening elements onto the stack
            stack.append((token, idx))
            prev_idx = None
        else:
            # Check that closing elements
            if not stack:
                raise UnbalancedException
            prev_open, prev_idx = stack.pop()
            want_close = open_to_close[prev_open]

            if token != want_close:
                raise UnbalancedException

        # If the stack is empty the sequence is currently balanced
        currently_balanced = not bool(stack)
        yield currently_balanced, token, prev_idx

    if stack:
        raise UnbalancedException


def balanced_decomp_unsafe(sequence, open_to_close):
    """
    Similar to :func:`balanced_decomp` but assumes that ``sequence`` is valid
    balanced sequence in order to execute faster.
    """
    gen = generate_balance_unsafe(sequence, open_to_close)

    bal_curr, tok_curr = next(gen)
    pop_open = sequence[0:1]
    want_close = open_to_close[tok_curr]

    head_stop = 1
    for head_stop, (bal_curr, tok_curr) in enumerate(gen, start=1):
        if bal_curr and tok_curr == want_close:
            pop_close = sequence[head_stop : head_stop + 1]
            break
    head = sequence[1:head_stop]
    tail = sequence[head_stop + 1 :]
    head_tail = head + tail
    return pop_open, pop_close, head, tail, head_tail


def generate_balance_unsafe(sequence, open_to_close):
    """
    Similar to :func:`generate_balance` but assumes that ``sequence`` is valid
    balanced sequence in order to execute faster.
    """
    stacklen = 0
    for token in sequence:
        if token in open_to_close:
            stacklen += 1
        else:
            stacklen -= 1
        yield stacklen == 0, token


def random_balanced_sequence(
    n, seed=None, item_type="chr", container_type="auto", open_to_close=None
):
    r"""
    Creates a random balanced sequence for testing / benchmarks

    Parameters
    ----------
    n : int
        A positive integer representing the number of nodes in the tree.

    seed : integer, random_state, or None (default)
        Indicator of random number generation state.
        See :ref:`Randomness<randomness>`.

    open_to_close : dict | None
        if specified, updates existing open_to_close with tokens from this
        sequence.

    item_type: str
        the type of sequence returned (see `item_type` in :func:`tree_to_seq`
        for details) can also be "paren", which is a special case that returns
        a nested set of parenthesis.

    container_type : str
        Determines the container_type type. Can be "auto", "list", "tuple", or
        "str". If "auto" tries to choose the best given the input data.

    Returns
    -------
    Tuple[(str | Tuple), Dict[str, str]]
        The first item is the sequence itself
        the second item is the open_to_close mappings.

    Example
    -------
    >>> # Demo the various sequence encodings that we might use
    >>> seq, open_to_close = random_balanced_sequence(4, seed=1, item_type="chr")
    >>> print("seq = {!r}".format(seq))
    seq = '\x00\x02\x04\x05\x03\x06\x07\x01'
    >>> seq, open_to_close = random_balanced_sequence(4, seed=1, item_type="number")
    >>> print("seq = {!r}".format(seq))
    seq = (1, 2, 3, -3, -2, 4, -4, -1)
    >>> seq, open_to_close = random_balanced_sequence(10, seed=1, item_type="paren")
    >>> print("seq = {!r}".format(seq))
    seq = '([[{{[]{[]}}{}()}]])'
    """
    from networkx_algo_common_subtree.tree_embedding import tree_to_seq
    from networkx_algo_common_subtree.utils import random_ordered_tree
    from networkx.utils import create_py_random_state

    # Create a random otree and then convert it to a balanced sequence
    rng = create_py_random_state(seed)

    if open_to_close is None:
        open_to_close = {}

    # To create a random balanced sequences we simply create a random ordered
    # tree and convert it to a sequence
    tree = random_ordered_tree(n, seed=rng, directed=True)
    if item_type == "paren":
        # special case
        pool = "[{("
        for node in tree.nodes:
            tree.nodes[node]["label"] = rng.choice(pool)
        open_to_close.update({"[": "]", "{": "}", "(": ")"})
        seq, open_to_close, *_ = tree_to_seq(
            tree,
            open_to_close=open_to_close,
            item_type="label",
            container_type=container_type,
        )
    else:
        seq, open_to_close, *_ = tree_to_seq(
            tree,
            open_to_close=open_to_close,
            item_type=item_type,
            container_type=container_type,
        )
    return seq, open_to_close



"""
Core python implementations for the longest common balanced sequence
subproblem, which is used by :mod:`tree_embedding`.
"""
import operator
from .balanced_sequence import (
    IdentityDict,
    generate_all_decomp,
    balanced_decomp_unsafe,
)

__all__ = [
    "available_impls_longest_common_balanced_embedding",
    "longest_common_balanced_embedding",
]


def longest_common_balanced_embedding(
    seq1, seq2, open_to_close, open_to_node=None, node_affinity="auto", impl="auto"
):
    """
    Finds the longest common balanced sequence embedding between two sequences

    This is used as a subproblem to solve the maximum common embedded subtree
    problem described in [1].

    Parameters
    ----------
    seq1, seq2: Sequence[TokT]
        two input balanced sequences

    open_to_close : Dict[TokT, TokT]
        a mapping from opening to closing tokens in the balanced sequence

    open_to_node : Dict[TokT, Any] | None
        If unspecified an identity mapping is assumed. Otherwise this is a
        dictionary that maps a sequence token to a value which is used in the
        ``node_affinity`` comparison.  Typically these are values corresponding
        to an original problem (e.g. a tree node).  This should only used in
        the case where the tokens in each sequence ``seq1`` and ``seq2`` are
        unique. NOTE: in the case where sequence tokens are not unique,
        sequences can always be re-encoded to differentiate between the same
        token at different indices without loss of generality.

    node_affinity : None | str | callable
        Function for to determine if two nodes can be matched. The function
        should take two arguments `node1` and `node2` and return a non-negative
        affinity score that is zero if the nodes are not allowed to match and
        some positive value indicating the strength of the match.  The return
        is interpreted as a weight that is used to break ties. If
        ``node_affinity=None`` then any node can match any other node and only
        the topology is important.  The default is "auto", which will use
        ``operator.eq`` to do a simple equality test on the nodes.

    impl : str
        Determines the backend implementation. Available choices are given by
        :func:`available_impls_longest_common_balanced_embedding`. The default
        is "auto", which chooses "iter-cython" if available, otherwise "iter".

    Returns
    -------
    Tuple[Tuple[Sequence[TokT], Sequence[TokT]], Float]
        A tuple indicating the common subsequence embedding of sequence1 and
        sequence2 (usually these are the same) and its value.

    See Also
    --------
    * This function is used to implement :func:`tree_embedding.maximum_common_ordered_subtree_embedding`

    Notes
    -----

    A balanced sequence is a sequence of tokens where there is a valid
    "nesting" of tokens given some relationship between opening and closing
    tokens (e.g. parenthesis, square brackets, and curly brackets). The
    following grammar generates all balanced sequences:

    .. code::

        t -> any opening token
        close(t) -> the closing token for t
        seq -> ''
        seq -> t + seq + close(t) + seq

    Given a balanced sequence s.

    .. code::

        s      = '([()[]])[{}([[]])]'

    We can use the grammar to decompose it as follows:

    .. code::

        s.a    = '('
        s.head =  '[()[]]'
        s.b    =        ')'
        s.tail =         '[{}([[]])]'

    Where s.a is the first token in s, and s.b is the corresponding closing
    token. s.head is everything between s.a and s.b, and s.tail is everything
    after s.b.

    Given two balanced sequences s1 and s2, their longest common subsequence
    embedding (lcse) is the largest common string you can obtain by deleting
    pairs of opening and closing tokens.

    We also define affinity as some degree of agreement between two tokens.
    By default we use an equality check: ``affinity(a1, a2) = (a1 == a2)``.

    The recurrence is defined as follows:

    .. code::

        lcse(s1, '') = 0
        lcse('', s2) = 0
        lcse(s1, s2) = max(
            lcse(s1.head, s2.head) + lcse(s1.tail, s2.tail) + affinity(s1.a, s2.a),
            lcse(s1.head + s1.tail, s2),
            lcse(s1, s2.head + s2.tail),
        )

    For example the longest common subsequence between the following s1 and s2
    are marked:

    .. code::

        seq1 = '([()[]])[{}([[]])]'
        _    = '        x   xxxx x'
        seq2 = '[[([])]]{[]}'
        _    = 'xx xx xx    '

        subseq = '[[[]]]'

    Also notice that '[[]]{}' is another solution to the previous example.
    That the longest common subsequence is not always unique.  For instance,
    consider the following two sequences:

    .. code::

        s1 = '({}[])'
        s2 = '[{}()]'

    They have three distinct longest common subsequences: '{}', '[]', and '()'.

    References
    ----------
    .. [1] Lozano, Antoni, and Gabriel Valiente.
        "On the maximum common embedded subtree problem for ordered trees."
        String Algorithmics (2004): 155-170.
        https://pdfs.semanticscholar.org/0b6e/061af02353f7d9b887f9a378be70be64d165.pdf

    Example
    -------
    >>> # Given two sequences and a mapping between opening and closing tokens
    >>> # we find the longest common subsequence (achievable by repeated
    >>> # balanced decomposition)
    >>> seq1 = "[][[]][]"
    >>> seq2 = "[[]][[]]"
    >>> open_to_close = {"[": "]"}
    >>> best, value = longest_common_balanced_embedding(seq1, seq2, open_to_close)
    ...
    >>> subseq1, subseq2 = best
    >>> print("subseq1 = {!r}".format(subseq1))
    subseq1 = '[][[]]'

    >>> # 1-label case from the paper (see Example 5)
    >>> # https://pdfs.semanticscholar.org/0b6e/061af02353f7d9b887f9a378be70be64d165.pdf
    >>> seq1 = "0010010010111100001011011011"
    >>> seq2 = "001000101101110001000100101110111011"
    >>> open_to_close = {"0": "1"}
    >>> best, value = longest_common_balanced_embedding(seq1, seq2, open_to_close)
    >>> subseq1, subseq2 = best
    >>> print("subseq1 = {!r}".format(subseq1))
    subseq1 = '00100101011100001011011011'
    >>> assert value == 13

    >>> # 3-label case
    >>> seq1 = "{({})([[]([]){(()(({()[]({}{})}))){}}])}"
    >>> seq2 = "{[({{}}{{[][{}]}(()[(({()})){[]()}])})]}"
    >>> open_to_close = {"{": "}", "(": ")", "[": "]"}
    >>> best, value = longest_common_balanced_embedding(seq1, seq2, open_to_close)
    >>> subseq1, subseq2 = best
    >>> print("subseq1 = {!r}".format(subseq1))
    subseq1 = '{{}[][]()(({()})){}}'
    >>> assert value == 10
    """
    if node_affinity == "auto" or node_affinity == "eq":
        node_affinity = operator.eq
    if node_affinity is None:

        def _matchany(a, b):
            return True

        node_affinity = _matchany
    if open_to_node is None:
        open_to_node = IdentityDict()
    full_seq1 = seq1
    full_seq2 = seq2
    if impl == "auto":
        if _cython_lcse_backend(error="ignore"):
            impl = "iter-cython"
        else:
            impl = "iter"

    if impl == "iter":
        value, best = _lcse_iter(
            full_seq1, full_seq2, open_to_close, node_affinity, open_to_node
        )
    elif impl == "iter-cython":
        balanced_embedding_cython = _cython_lcse_backend(error="raise")
        value, best = balanced_embedding_cython._lcse_iter_cython(
            full_seq1, full_seq2, open_to_close, node_affinity, open_to_node
        )
    elif impl == "recurse":
        _memo = {}
        _seq_memo = {}
        value, best = _lcse_recurse(
            full_seq1,
            full_seq2,
            open_to_close,
            node_affinity,
            open_to_node,
            _memo,
            _seq_memo,
        )
    else:
        raise KeyError(impl)
    return best, value


def available_impls_longest_common_balanced_embedding():
    """
    Returns all available implementations for
    :func:`longest_common_balanced_embedding`.

    Returns
    -------
    List[str]
        the string code for each available implementation
    """
    impls = []
    if _cython_lcse_backend():
        impls += [
            "iter-cython",
        ]

    # Pure python backends
    impls += [
        "iter",
        "recurse",
    ]
    return impls


def _cython_lcse_backend(error="ignore", verbose=0):
    """
    Returns the cython backend if available, otherwise None

    CommandLine
    -----------
    xdoctest -m balanced_embedding _cython_lcse_backend
    """
    try:
        # Attempt to use the module build with CMake
        from . import balanced_embedding_cython
    except Exception:
        if error == "ignore":
            balanced_embedding_cython = None
        elif error == "raise":
            raise
        else:
            raise KeyError(error)
    return balanced_embedding_cython


def _lcse_iter(full_seq1, full_seq2, open_to_close, node_affinity, open_to_node):
    """
    Depth first stack trajectory and replace try except statements with ifs

    This is the current best pure-python algorithm candidate

    Converts :func:`_lcse_recurse` into an iterative algorithm using a fairly
    straightforward method that effectively simulates callstacks.  Uses a
    breadth-first trajectory and try-except to catch missing memoized results
    (which seems to be slightly faster than if statements).

    Example
    -------
    >>> full_seq1 = "[][[]][]"
    >>> full_seq2 = "[[]][[]]"
    >>> open_to_close = {"[": "]"}
    >>> import operator as op
    >>> node_affinity = op.eq
    >>> open_to_node = IdentityDict()
    >>> res = _lcse_iter(full_seq1, full_seq2, open_to_close, node_affinity,
    ...                 open_to_node)
    >>> val, embeddings = res
    >>> print(embeddings[0])
    [][[]]
    """
    all_decomp1 = generate_all_decomp(full_seq1, open_to_close, open_to_node)
    all_decomp2 = generate_all_decomp(full_seq2, open_to_close, open_to_node)

    key0 = (full_seq1, full_seq2)
    frame0 = key0
    stack = [frame0]

    # Memoize mapping (seq1, seq2) -> best size, embeddings
    _results = {}

    # Populate base cases
    empty1 = type(next(iter(all_decomp1.keys())))()
    empty2 = type(next(iter(all_decomp2.keys())))()
    best = (empty1, empty2)
    base_result = (0, best)
    for seq1 in all_decomp1.keys():
        key1 = seq1
        t1, a1, b1, head1, tail1, head_tail1 = all_decomp1[key1]
        _results[(seq1, empty2)] = base_result
        _results[(head1, empty2)] = base_result
        _results[(tail1, empty2)] = base_result
        _results[(head_tail1, empty2)] = base_result

    for seq2 in all_decomp2.keys():
        key2 = seq2
        t2, a2, b2, head2, tail2, head_tail2 = all_decomp2[key2]
        _results[(empty1, seq2)] = base_result
        _results[(empty1, head2)] = base_result
        _results[(empty1, tail2)] = base_result
        _results[(empty1, head_tail2)] = base_result

    while stack:
        key = stack[-1]
        if key not in _results:
            seq1, seq2 = key

            t1, a1, b1, head1, tail1, head_tail1 = all_decomp1[seq1]
            t2, a2, b2, head2, tail2, head_tail2 = all_decomp2[seq2]

            # Case 2: The current edge in sequence1 is deleted
            try_key = (head_tail1, seq2)
            if try_key in _results:
                cand1 = _results[try_key]
            else:
                stack.append(try_key)
                continue

            # Case 3: The current edge in sequence2 is deleted
            try_key = (seq1, head_tail2)
            if try_key in _results:
                cand2 = _results[try_key]
            else:
                stack.append(try_key)
                continue

            # Case 1: The LCSE involves this edge
            affinity = node_affinity(t1, t2)
            if affinity:
                try_key = (head1, head2)
                if try_key in _results:
                    pval_h, new_heads = _results[try_key]
                else:
                    stack.append(try_key)
                    continue

                try_key = (tail1, tail2)
                if try_key in _results:
                    pval_t, new_tails = _results[try_key]
                else:
                    stack.append(try_key)
                    continue

                new_head1, new_head2 = new_heads
                new_tail1, new_tail2 = new_tails

                subseq1 = a1 + new_head1 + b1 + new_tail1
                subseq2 = a2 + new_head2 + b2 + new_tail2

                res3 = (subseq1, subseq2)
                val3 = pval_h + pval_t + affinity
                cand3 = (val3, res3)
            else:
                cand3 = (-1, None)

            # We solved the frame
            _results[key] = max(cand1, cand2, cand3)
        stack.pop()

    found = _results[key0]
    return found


def _lcse_recurse(
    seq1, seq2, open_to_close, node_affinity, open_to_node, _memo, _seq_memo
):
    """
    Surprisingly, this recursive implementation is one of the faster
    pure-python methods for certain input types. However, its major drawback is
    that it can raise a RecursionError if the inputs are too deep.

    See also the iterative version :func:`_lcse_iter`
    """
    if not seq1:
        return 0, (seq1, seq1)
    elif not seq2:
        return 0, (seq2, seq2)
    else:
        key1 = hash(seq1)  # using hash(seq) is faster than seq itself
        key2 = hash(seq2)
        key = hash((key1, key2))
        if key in _memo:
            return _memo[key]

        if key1 in _seq_memo:
            a1, b1, head1, tail1, head1_tail1 = _seq_memo[key1]
        else:
            a1, b1, head1, tail1, head1_tail1 = balanced_decomp_unsafe(
                seq1, open_to_close
            )
            _seq_memo[key1] = a1, b1, head1, tail1, head1_tail1

        if key2 in _seq_memo:
            a2, b2, head2, tail2, head2_tail2 = _seq_memo[key2]
        else:
            a2, b2, head2, tail2, head2_tail2 = balanced_decomp_unsafe(
                seq2, open_to_close
            )
            _seq_memo[key2] = a2, b2, head2, tail2, head2_tail2

        # Case 2: The current edge in sequence1 is deleted
        val, best = _lcse_recurse(
            head1_tail1,
            seq2,
            open_to_close,
            node_affinity,
            open_to_node,
            _memo,
            _seq_memo,
        )

        # Case 3: The current edge in sequence2 is deleted
        val_alt, cand = _lcse_recurse(
            seq1,
            head2_tail2,
            open_to_close,
            node_affinity,
            open_to_node,
            _memo,
            _seq_memo,
        )
        if val_alt > val:
            best = cand
            val = val_alt

        # Case 1: The LCSE involves this edge
        t1 = open_to_node[a1[0]]
        t2 = open_to_node[a2[0]]
        affinity = node_affinity(t1, t2)
        if affinity:
            pval_h, new_heads = _lcse_recurse(
                head1,
                head2,
                open_to_close,
                node_affinity,
                open_to_node,
                _memo,
                _seq_memo,
            )
            pval_t, new_tails = _lcse_recurse(
                tail1,
                tail2,
                open_to_close,
                node_affinity,
                open_to_node,
                _memo,
                _seq_memo,
            )

            new_head1, new_head2 = new_heads
            new_tail1, new_tail2 = new_tails

            subseq1 = a1 + new_head1 + b1 + new_tail1
            subseq2 = a2 + new_head2 + b2 + new_tail2

            cand = (subseq1, subseq2)
            val_alt = pval_h + pval_t + affinity
            if val_alt > val:
                best = cand
                val = val_alt

        found = (val, best)
        _memo[key] = found
        return found



import networkx as nx
import textwrap
from networkx.utils import create_py_random_state
from networkx_algo_common_subtree.balanced_isomorphism import (
    available_impls_longest_common_balanced_isomorphism,
)
from networkx_algo_common_subtree import (
    maximum_common_ordered_subtree_isomorphism,
    maximum_common_ordered_subtree_embedding
)
from networkx_algo_common_subtree.utils import random_ordered_tree
from networkx_algo_common_subtree.utils import graph_str
from networkx_algo_common_subtree._types import OrderedDiGraph
from networkx import isomorphism


def test_all_implementations_are_same():
    """
    Tests several random sequences
    """

    seed = 24658885408229410362279507020239
    rng = create_py_random_state(seed)

    maxsize = 20
    num_trials = 30

    node_affinity = "eq"

    for _ in range(num_trials):
        n1 = rng.randint(1, maxsize)
        n2 = rng.randint(1, maxsize)

        tree1 = random_ordered_tree(n1, seed=rng, directed=True)
        tree2 = random_ordered_tree(n2, seed=rng, directed=True)

        # Note: the returned sequences may be different (maximum embeddings may
        # not be unique), but the values should all be the same.
        affinity_funcs = ["eq", None]
        for node_affinity in affinity_funcs:
            results = {}
            impls = available_impls_longest_common_balanced_isomorphism()
            for impl in impls:
                # FIXME: do we need to rework the return value here?
                subtree1, subtree2, _ = maximum_common_ordered_subtree_isomorphism(
                    tree1, tree2, node_affinity=node_affinity, impl=impl
                )
                _check_subtree_isomorphism_invariants(
                    tree1, tree2, subtree1, subtree2, node_affinity
                )
                results[impl] = len(subtree1.nodes)

            x = max(results.values())
            assert all(v == x for v in results.values())


def check_btree_case():
    tree1 = nx.balanced_tree(3, 2, create_using=OrderedDiGraph)
    tree2 = nx.balanced_tree(2, 2, create_using=OrderedDiGraph)

    nx.set_node_attributes(
        tree1, dict(zip(tree1.nodes, map(str, tree1.nodes))), "label"
    )
    nx.set_node_attributes(
        tree2, dict(zip(tree2.nodes, map(str, tree2.nodes))), "label"
    )
    # tree1.remove_node(6)
    # tree1.remove_node(7)

    node_affinity = "eq"

    item_type = "auto"
    impl = "auto"

    node_affinity = "eq"

    subtree1, subtree2, _ = maximum_common_ordered_subtree_isomorphism(
        tree1, tree2, node_affinity=node_affinity, item_type=item_type, impl=impl
    )
    nx.set_node_attributes(
        subtree1, dict(zip(subtree1.nodes, map(str, subtree1.nodes))), "label"
    )
    nx.set_node_attributes(
        subtree2, dict(zip(subtree2.nodes, map(str, subtree2.nodes))), "label"
    )

    embedding1, embedding2, _ = maximum_common_ordered_subtree_embedding(
        tree1,
        tree2,
        node_affinity=node_affinity,
    )

    def node_match(n1, n2):
        return n1 == n2

    matcher = isomorphism.DiGraphMatcher(tree1, tree2, node_match=node_match)
    print(list(matcher.isomorphisms_iter()))
    print(list(matcher.subgraph_isomorphisms_iter()))
    print(list(matcher.subgraph_monomorphisms_iter()))
    print(matcher.subgraph_is_isomorphic())

    matcher = isomorphism.DiGraphMatcher(tree1, subtree2, node_match=node_match)
    print(list(matcher.isomorphisms_iter()))
    print(list(matcher.subgraph_isomorphisms_iter()))
    print(list(matcher.subgraph_monomorphisms_iter()))
    print(matcher.subgraph_is_isomorphic())

    verbose = 1
    if verbose:
        print("tree1")
        print(graph_str(tree1))
        print("tree2")
        print(graph_str(tree2))

        print("subtree1")
        print(graph_str(subtree1))
        print("subtree2")
        print(graph_str(subtree2))

        print("embedding1")
        print(graph_str(embedding1))
        print("embedding2")
        print(graph_str(embedding2))


def check_dfs_relabel_case():
    tree1 = nx.balanced_tree(2, 4, create_using=OrderedDiGraph)
    tree2 = nx.balanced_tree(3, 4, create_using=OrderedDiGraph)
    # Relabel trees so DFS nodes in the first path match
    mapping1 = dict(zip(nx.dfs_preorder_nodes(tree1), range(len(tree1))))
    mapping2 = dict(zip(nx.dfs_preorder_nodes(tree2), range(len(tree2))))
    tree1 = nx.relabel_nodes(tree1, mapping1)
    tree2 = nx.relabel_nodes(tree2, mapping2)
    print("tree1")
    print(graph_str(tree1))
    print("tree2")
    print(graph_str(tree2))

    subtree1, subtree2, _ = maximum_common_ordered_subtree_isomorphism(tree1, tree2)
    print("subtree1")
    print(graph_str(subtree1))
    print("subtree2")
    print(graph_str(subtree2))
    assert set(subtree1.nodes) == set(range(6))


def check_any_match_case():
    # Allow any node to match any node. The result should be an r2-h2 btree
    tree1 = nx.balanced_tree(2, 3, create_using=OrderedDiGraph)
    tree2 = nx.balanced_tree(3, 2, create_using=OrderedDiGraph)
    print("tree1")
    print(graph_str(tree1))
    print("tree2")
    print(graph_str(tree2))
    subtree1, subtree2, _ = maximum_common_ordered_subtree_isomorphism(
        tree1, tree2, node_affinity=None
    )
    print("subtree1")
    print(graph_str(subtree1))
    print("subtree2")
    print(graph_str(subtree2))
    assert set(subtree1.nodes) == set(range(6))

    expected = nx.balanced_tree(2, 2, create_using=OrderedDiGraph)
    # print(graph_str(expected))
    assert isomorphism.is_isomorphic(subtree1, expected)
    assert isomorphism.is_isomorphic(subtree2, expected)

    matcher = isomorphism.DiGraphMatcher(
        tree1, subtree2, node_match=lambda n1, n2: True
    )
    assert matcher.subgraph_is_isomorphic()

    matcher = isomorphism.DiGraphMatcher(
        tree2, subtree1, node_match=lambda n1, n2: True
    )
    assert matcher.subgraph_is_isomorphic()
    list(matcher.subgraph_isomorphisms_iter())


def check_simple_skip_head_case():
    tree1 = OrderedDiGraph()
    tree1.add_edge(0, 1)
    tree1.add_edge(1, 4)
    tree1.add_edge(1, 5)

    tree2 = OrderedDiGraph()
    tree2.add_edge(0, 1)
    tree2.add_edge(1, 3)
    tree2.add_edge(1, 4)

    item_type = "number"

    subtree1, subtree2, _ = maximum_common_ordered_subtree_isomorphism(
        tree1, tree2, item_type=item_type
    )

    print("tree1")
    print(graph_str(tree1))
    print("tree2")
    print(graph_str(tree2))

    print("subtree1")
    print(graph_str(subtree1))
    print("subtree2")
    print(graph_str(subtree2))


def _check_subtree_isomorphism_invariants(
    tree1, tree2, subtree1, subtree2, node_affinity
):

    # Subtrees should already be isomorphic
    assert isomorphism.is_isomorphic(subtree1, subtree2)

    if len(subtree1) == 0:
        return

    # Maximum isomorphism should be the same between already found subtrees
    alt_subtree1, alt_subtree2, _ = maximum_common_ordered_subtree_isomorphism(
        subtree1, subtree2, node_affinity=node_affinity
    )
    assert set(alt_subtree1.edges) == set(subtree1.edges)
    assert set(alt_subtree2.edges) == set(subtree2.edges)

    embedding1, embedding2, _ = maximum_common_ordered_subtree_embedding(
        tree1, tree2, node_affinity=node_affinity
    )
    assert len(embedding1) >= len(
        subtree1
    ), "embeddings should at least as big as induced-isomorphisms"
    assert len(embedding2) >= len(
        subtree2
    ), "embeddings should at least as big as induced-isomorphisms"

    assert isomorphism.DiGraphMatcher(tree1, subtree1).subgraph_is_isomorphic()
    assert isomorphism.DiGraphMatcher(tree2, subtree2).subgraph_is_isomorphic()

    alt_subtree1, _, _ = maximum_common_ordered_subtree_isomorphism(
        subtree1, tree2, node_affinity=node_affinity
    )
    _, alt_subtree2, _ = maximum_common_ordered_subtree_isomorphism(
        tree1, subtree2, node_affinity=node_affinity
    )
    assert set(alt_subtree1.edges) == set(subtree1.edges)
    assert set(alt_subtree2.edges) == set(subtree2.edges)

    alt_embed1, _, _ = maximum_common_ordered_subtree_embedding(
        subtree1, tree2, node_affinity=node_affinity
    )
    _, alt_embed2, _ = maximum_common_ordered_subtree_embedding(
        tree1, subtree2, node_affinity=node_affinity
    )
    assert set(alt_embed1.edges) == set(subtree1.edges)
    assert set(alt_embed2.edges) == set(subtree2.edges)

    verbose = 0
    if verbose:
        print("tree1")
        print(graph_str(tree1))
        print("tree2")
        print(graph_str(tree2))

        print("subtree1")
        print(graph_str(subtree1))
        print("subtree2")
        print(graph_str(subtree2))


def test_custom_large_case():
    # Create two trees that have a reasonably complex shared structure
    from networkx_algo_common_subtree.utils import random_ordered_tree
    tree1 = random_ordered_tree(10, seed=3)
    tree2 = random_ordered_tree(10, seed=2)
    tree1.add_edges_from(tree2.edges, weight=1)
    tree1 = nx.minimum_spanning_arborescence(tree1)
    tree2.add_edges_from(tree1.edges, weight=1)
    tree2 = nx.minimum_spanning_arborescence(tree2)

    tree1.remove_edge(4, 7)
    tree1.remove_edge(4, 9)
    tree1.add_edge(4, 10)
    tree1.add_edge(10, 7)
    tree1.add_edge(10, 9)
    tree1.add_edges_from([(9, 11), (11, 12)])
    tree2.add_edges_from([(9, 11), (11, 12)])
    tree2.add_edge(100, 0)
    tree1.add_edge(102, 100)
    tree1.add_edge(100, 101)
    tree1.add_edge(101, 0)
    tree1.add_edge(5, 201)
    tree1.add_edge(5, 202)
    tree1.add_edge(5, 203)
    tree1.add_edge(201, 2000)
    tree1.add_edge(2000, 2001)
    tree1.add_edge(2001, 2002)
    tree1.add_edge(2002, 2003)

    tree2.add_edge(5, 202)
    tree2.add_edge(5, 203)
    tree2.add_edge(5, 201)
    tree2.add_edge(201, 2000)
    tree2.add_edge(2000, 2001)
    tree2.add_edge(2001, 2002)
    tree2.add_edge(2002, 2003)

    print("-----")
    print("tree1")
    print(graph_str(tree1))
    print("tree2")
    print(graph_str(tree2))

    subtree1, subtree2, weight = maximum_common_ordered_subtree_isomorphism(
        tree1, tree2
    )
    subtree1_got = graph_str(subtree1)
    subtree2_got = graph_str(subtree2)
    target = textwrap.dedent(
        """
        ╙── 0
            ├─╼ 3
            │   └─╼ 8
            └─╼ 1
                ├─╼ 6
                └─╼ 5
                    ├─╼ 2
                    │   └─╼ 4
                    └─╼ 201
                        └─╼ 2000
                            └─╼ 2001
                                └─╼ 2002
                                    └─╼ 2003
        """
    ).strip()
    print("-----")
    print("subtree1")
    print(subtree1_got)
    print("subtree2")
    print(subtree2_got)

    assert target == subtree1_got
    assert target == subtree2_got

    # Compare the subtree isomorphism to the subtree embedding
    embedding1, embedding2, _ = maximum_common_ordered_subtree_embedding(
        tree1, tree2
    )
    print("-----")
    print("embedding1")
    print(graph_str(embedding1))
    print("embedding2")
    print(graph_str(embedding2))

    if 0:
        import timerit

        ti = timerit.Timerit(6, bestof=2, verbose=2)
        for timer in ti.reset("isomorphism"):
            with timer:
                maximum_common_ordered_subtree_isomorphism(tree1, tree2)
        for timer in ti.reset("embedding"):
            with timer:
                maximum_common_ordered_subtree_embedding(tree1, tree2)

    assert isomorphism.DiGraphMatcher(tree1, subtree1).subgraph_is_isomorphic()
    assert isomorphism.DiGraphMatcher(tree2, subtree2).subgraph_is_isomorphic()

    if 0:
        list(isomorphism.DiGraphMatcher(tree1, tree2).subgraph_isomorphisms_iter())
        list(isomorphism.DiGraphMatcher(tree1, tree2).subgraph_monomorphisms_iter())

        list(
            isomorphism.DiGraphMatcher(subtree1, subtree2).subgraph_isomorphisms_iter()
        )
        list(isomorphism.DiGraphMatcher(tree1, subtree1).subgraph_isomorphisms_iter())
        list(isomorphism.DiGraphMatcher(tree2, subtree2).subgraph_isomorphisms_iter())



"""
Algorithm for computing the largest common tree embeddings (also known as a
minor) shared by two trees. See :func:`maximum_common_ordered_subtree_embedding`
for more details.
"""
from . import balanced_sequence
from . import balanced_embedding
from ._types import OrderedDiGraph
import networkx as nx

__all__ = ["maximum_common_ordered_subtree_embedding"]


def maximum_common_ordered_subtree_embedding(
    tree1, tree2, node_affinity="auto", impl="auto", item_type="auto"
):
    r"""
    Finds the maximum common subtree-embedding between two ordered trees.

    A tree S is an embedded subtree (also known as a minor) of T if it can be
    obtained from T by a series of edge contractions.

    Subtree embeddings (or minors) are similar to tree isomorphisms --- if T is
    a subtree isomorphism then T is a minor. However, if you contract an edge in
    T it, then it may no longer be an isomorphism, but it is still a minor.

    This function computes the maximum common embedded subtrees S1 and S2
    between two trees T1 and T2. S1 and S2 are minors of T1 and T2 with maximal
    size such that S1 is isomorphic to S2.

    The computational complexity is: ``O(n1 * n2 * min(d1, l1) * min(d2, l2))``
    on ordered trees with n1 and n2 nodes, of depth d1 and d2 and with l1 and
    l2 leaves, respectively.

    This implementation follows the algorithm described in [1]_, which
    introduces the problem as follows:

    "An important generalization of tree and subtree isomorphism, known as
    minor containment, is the problem of determining whether a tree is
    isomorphic to an embedded subtree of another tree, where an embedded
    subtree of a tree is obtained by contracting some of the edges in the tree.
    A further generalization of minor containment on trees, known as maximum
    common embedded subtree, is the problem of finding or determining the size
    of a largest common embedded subtree of two trees. The latter also
    generalizes the maximum common subtree isomorphism problem, in which a
    common subtree of largest size is contained as a subtree, not only
    embedded, in the two trees."

    Parameters
    ----------
    tree1, tree2 : OrderedDiGraph
        Trees to find the maximum embedding between

    node_affinity : None | str | callable
        Function for to determine if two nodes can be matched. The return is
        interpreted as a weight that is used to break ties. If None then any
        node can match any other node and only the topology is important.
        The default is "eq", which is the same as ``operator.eq``.

    impl : str
        Determines the backend implementation. Defaults to "auto".
        See :func:`balanced_embedding.longest_common_balanced_embedding`
        for details. Other valid options are "iter", "recurse", and
        "iter-cython".

    item_type : str
        Determines the backend data structure used to encode the tree as a
        balanced sequence. Defaults to "auto", other valid options are "chr"
        and "number".

    Returns
    -------
    S1, S2, value: Tuple[OrderedDiGraph, OrderedDiGraph, float]
        The maximum value common embedding for each tree with respect to the
        chosen ``node_affinity`` function. The topology of both graphs will
        always be the same, the only difference is that the node labels in the
        first and second embeddings will correspond to ``tree1`` and ``tree2``
        respectively. When ``node_affinity='eq'`` then embeddings should be
        identical. The last return value is the "weight" of the solution with
        respect to ``node_affinity``.

    References
    ----------
    .. [1] Lozano, Antoni, and Gabriel Valiente.
        "On the maximum common embedded subtree problem for ordered trees."
        String Algorithmics (2004): 155-170.
        https://pdfs.semanticscholar.org/0b6e/061af02353f7d9b887f9a378be70be64d165.pdf

    See Also
    --------
    * For example usage see ``examples/path_correspondence.py``
    * Core backends are in :mod:`balanced_embedding.longest_common_balanced_embedding`

    Example
    -------
    >>> # Create two random trees
    >>> from networkx_algo_common_subtree.utils import random_ordered_tree
    >>> from networkx_algo_common_subtree.utils import graph_str
    >>> tree1 = random_ordered_tree(7, seed=3257073545741117277206611, directed=True)
    >>> tree2 = random_ordered_tree(7, seed=123568587133124688238689717, directed=True)
    >>> print(graph_str(tree1))
    ╙── 0
        ├─╼ 5
        │   └─╼ 2
        └─╼ 1
            └─╼ 6
                ├─╼ 3
                └─╼ 4
    >>> print(graph_str(tree2))
    ╙── 0
        └─╼ 2
            ├─╼ 1
            │   ├─╼ 4
            │   └─╼ 3
            │       └─╼ 5
            └─╼ 6
    >>> # Compute the maximum common embedding between the two trees
    >>> embedding1, embedding2, _ = maximum_common_ordered_subtree_embedding(tree1, tree2)
    >>> print(graph_str(embedding1))
    ╙── 0
        └─╼ 1
            └─╼ 4
    >>> assert embedding1.edges == embedding2.edges, (
    ...     'when node_affinity is "eq" both embeddings will be the same')

    >>> # Demo with a custom node affinity where any node can match unless
    >>> # they are the same and we much prefer nodes that are disimilar
    >>> def custom_node_affinity(n1, n2):
    ...     return abs(n1 - n2) ** 2
    >>> embedding1, embedding2, _ = maximum_common_ordered_subtree_embedding(
    ...     tree1, tree2, node_affinity=custom_node_affinity)
    >>> # In this case the embeddings for each tree will be differnt
    >>> print(graph_str(embedding1))
    ╙── 0
        ├─╼ 5
        │   └─╼ 2
        └─╼ 1
    >>> print(graph_str(embedding2))
    ╙── 2
        ├─╼ 1
        │   └─╼ 5
        └─╼ 6
    """
    # Note: checks that inputs are forests are handled by tree_to_seq
    if not isinstance(tree1, OrderedDiGraph):
        raise nx.NetworkXNotImplemented(
            "only implemented for directed ordered trees. "
            "Got {} instead".format(type(tree1))
        )
    if not isinstance(tree1, OrderedDiGraph):
        raise nx.NetworkXNotImplemented(
            "only implemented for directed ordered trees. "
            "Got {} instead".format(type(tree2))
        )

    if tree1.number_of_nodes() == 0 or tree2.number_of_nodes() == 0:
        raise nx.NetworkXPointlessConcept

    if item_type == "label":
        # If we do allow label, I think the algorithm will work, but the
        # returned tree embeddings will only be embedding wrt to the label
        # structure.
        raise AssertionError(
            "allowing sequences to be specified by the labels breaks assumptions"
        )

    # Convert the trees to balanced sequences.
    # NOTE: each sequence will contain each token at most once, this is an
    # important assumption in subsequent steps.
    seq1, open_to_close, node_to_open = tree_to_seq(
        tree1,
        open_to_close=None,
        node_to_open=None,
        item_type=item_type,
        container_type="auto",
    )
    seq2, open_to_close, node_to_open = tree_to_seq(
        tree2, open_to_close, node_to_open, item_type=item_type, container_type="auto"
    )

    # NOTE: This DOES work in the case where all opening tokens within a single
    # sequence are unique. And we CAN enforce that this is the case in our
    # reduction because each node in a graph is always unique and we always
    # choose a unique token for each unique node in ``tree_to_seq``.
    open_to_node = {tok: node for node, tok in node_to_open.items()}

    # Solve the longest common balanced sequence problem
    best, value = balanced_embedding.longest_common_balanced_embedding(
        seq1,
        seq2,
        open_to_close,
        open_to_node=open_to_node,
        node_affinity=node_affinity,
        impl=impl,
    )
    subseq1, subseq2 = best

    # Convert the subsequence back into a tree.
    # Note: we could return the contracted edges as well here, but that can
    # always be done as a postprocessing step. See tests for an example of
    # this.
    embedding1 = seq_to_tree(subseq1, open_to_close, open_to_node)
    embedding2 = seq_to_tree(subseq2, open_to_close, open_to_node)

    return embedding1, embedding2, value


def tree_to_seq(
    tree, open_to_close=None, node_to_open=None, item_type="auto", container_type="auto"
):
    r"""
    Converts an ordered tree to a balanced sequence --- typically with unique
    tokens --- for use in algorithm reductions.

    Used to convert a tree to a sequence before solving
    :func:`longest_common_balanced_embedding` in
    :func:`maximum_common_ordered_subtree_embedding`.

    Parameters
    ----------
    tree: OrderedDiGraph
        The forest to encode as a string sequence.

    open_to_close : Dict | None
        Dictionary of opening to closing tokens to be updated for problems
        where multiple trees are converted to sequences.

    node_to_open : Dict | None
        Dictionary of nodes mapped to the opening tokens to be updated for
        problems where multiple trees are converted to sequences.

    item_type : str
        Determines the item type of the sequence.
        Can be 'auto', 'number', 'chr', or 'label'.
        Default is 'auto', which will choose 'chr' if the graph is small enough
        otherwise 'number'.  If item_type is 'label', then the label of each
        node is used to create the token, and the `open_to_close` dictionary
        must be specified.

    container_type : str
        Determines the container_type type. Can be "auto", "list", "tuple", or
        "str". If "auto" tries to choose the best given the input data.

    Returns:
    --------
    Tuple[SeqT, Dict, Dict]
        A tuple containing
            sequence - the string representation of an ordered tree
            open_to_close - a mapping between opening and closing tokens
            node_to_open - a mapping between tree nodes and opening tokens

    Examples
    --------
    >>> # This function helps us encode this graph as a balance sequence
    >>> from networkx_algo_common_subtree.utils import graph_str
    >>> tree = nx.path_graph(3, OrderedDiGraph)
    >>> print(graph_str(tree))
    ╙── 0
        └─╼ 1
            └─╼ 2
    >>> # The sequence is represented by opening and closing tokens
    >>> # These are returned a container, which might be a tuple of numbers
    >>> sequence, open_to_close, node_to_open, *_ = tree_to_seq(tree, item_type='number')
    >>> print(('''
    ... sequence = {sequence}
    ... open_to_close = {open_to_close}
    ... node_to_open = {node_to_open}
    ... ''').format(**locals()).strip())
    sequence = (1, 2, 3, -3, -2, -1)
    open_to_close = {1: -1, 2: -2, 3: -3}
    node_to_open = {0: 1, 1: 2, 2: 3}

    >>> # But you might also encode as a sequence of utf8-characters
    >>> # These can often be quicker to use than number encodings
    >>> sequence, open_to_close, node_to_open, *_ = tree_to_seq(tree, item_type='chr')
    >>> print(('''
    ... sequence = {sequence!r}
    ... open_to_close = {open_to_close!r}
    ... node_to_open = {node_to_open!r}
    ... ''').format(**locals()).strip())
    sequence = '\x00\x02\x04\x05\x03\x01'
    open_to_close = {'\x00': '\x01', '\x02': '\x03', '\x04': '\x05'}
    node_to_open = {0: '\x00', 1: '\x02', 2: '\x04'}

    >>> # Here is a more complex example
    >>> tree = nx.balanced_tree(2, 2, nx.DiGraph)
    >>> print(graph_str(tree))
    ╙── 0
        ├─╼ 1
        │   ├─╼ 3
        │   └─╼ 4
        └─╼ 2
            ├─╼ 5
            └─╼ 6
    >>> sequence, *_ = tree_to_seq(tree, item_type='number')
    >>> print('sequence = {!r}'.format(sequence))
    sequence = (1, 2, 3, -3, 4, -4, -2, 5, 6, -6, 7, -7, -5, -1)
    >>> sequence, *_ = tree_to_seq(tree, item_type='chr')
    >>> print('sequence = {!r}'.format(sequence))
    sequence = '\x00\x02\x04\x05\x06\x07\x03\x08\n\x0b\x0c\r\t\x01'

    >>> # Demo custom label encoding: If you have custom labels on your
    >>> # tree nodes, those can be used in the encoding.
    >>> from networkx_algo_common_subtree.utils import random_ordered_tree
    >>> import random
    >>> tree = random_ordered_tree(10, seed=1, directed=True)
    >>> rng = random.Random(0)
    >>> open_to_close = dict(zip("[{(", "]})"))
    >>> for node in tree.nodes:
    ...     tree.nodes[node]["label"] = rng.choice(list(open_to_close.keys()))
    >>> sequence, *_ = tree_to_seq(tree, item_type="label", container_type="str", open_to_close=open_to_close)
    >>> print('sequence = {!r}'.format(sequence))
    sequence = '{[{{{{}({})}{}{}}}]}'
    """
    # Create a sequence and mapping from each index in the sequence to the
    # graph edge is corresponds to.
    sequence = []

    # mapping between opening and closing tokens
    if open_to_close is None:
        open_to_close = {}
    if node_to_open is None:
        node_to_open = {}

    # utf8 can only encode this many chars
    NUM_CHRS = 1112064
    NUM_OPEN_CHRS = NUM_CHRS // 2

    if item_type == "label":
        # Special case, where the user specifies the encoding
        all_labels = {n["label"] for n in tree.nodes.values()}

        if container_type in {"auto", "str"}:
            # Determine if the container_type can be a string
            can_be_str = all(isinstance(x, str) and len(x) == 1 for x in all_labels)
            if container_type == "str" and not can_be_str:
                raise ValueError("Labels cannot be contained as a string")
            if container_type == "auto":
                container_type = "str" if can_be_str else "tuple"

        if not open_to_close:
            raise ValueError("must specify open_to_close for custom labeling")
    else:
        # Normal case where we will define the sequence encoding for the tree
        if item_type == "auto":
            # chr mode is fastest but limited to ~half-a-million nodes
            item_type = "chr" if len(tree) < NUM_OPEN_CHRS else "number"
        if container_type == "auto":
            container_type = "str" if item_type == "chr" else "tuple"

    sources = [n for n in tree.nodes if tree.in_degree[n] == 0]
    dfs_forest_edge_gen = (
        (u, v, etype)
        for source in sources
        for u, v, etype in nx.dfs_labeled_edges(tree, source=source)
    )
    for u, v, etype in dfs_forest_edge_gen:
        if etype == "forward":
            # u has been visited by v has not
            if v not in node_to_open:
                if item_type == "number":
                    # Pos nums are open toks. Neg nums are close toks.
                    open_tok = len(node_to_open) + 1
                    close_tok = -open_tok
                elif item_type == "chr":
                    # Even chars are open toks. Odd chars are close toks.
                    open_tok = chr(len(node_to_open) * 2)
                    close_tok = chr(len(node_to_open) * 2 + 1)
                elif item_type == "label":
                    # The user must specify the closing token
                    open_tok = tree.nodes[v]["label"]
                    close_tok = open_to_close[open_tok]
                else:
                    raise KeyError(item_type)
                node_to_open[v] = open_tok
                open_to_close[open_tok] = close_tok
            open_tok = node_to_open[v]
            sequence.append(open_tok)
        elif etype == "reverse":
            # Both u and v are visited and the edge is in the tree
            close_tok = open_to_close[node_to_open[v]]
            sequence.append(close_tok)
        elif etype == "nontree":
            raise TypeError("Input must be a forest")
        else:
            raise KeyError(etype)

    if item_type == "chr":
        assert len(node_to_open) < NUM_OPEN_CHRS, "graph is way too big"

    if container_type == "str":
        sequence = "".join(sequence)
    elif container_type == "list":
        sequence = sequence
    elif container_type == "tuple":
        sequence = tuple(sequence)
    else:
        raise KeyError(container_type)

    return sequence, open_to_close, node_to_open


def seq_to_tree(subseq, open_to_close, open_to_node):
    """
    Converts a balanced sequence to an ordered tree

    Used to convert back to a tree after solving
    :func:`longest_common_balanced_embedding` in
    :func:`maximum_common_ordered_subtree_embedding`.

    Parameters
    ----------
    subseq : Tuple | str
        a balanced sequence of hashable items as a string or tuple

    open_to_close : Dict
        a dictionary that maps opening tokens to closing tokens in the balanced
        sequence problem.

    open_to_node : Dict
        a dictionary that maps a sequence token to a node corresponding to an
        original problem (e.g. a tree node). Must be unique. If unspecified new
        nodes will be generated and the opening sequence token will be used as
        a node label.

    Returns
    -------
    subtree: OrderedDiGraph
        The ordered tree that corresponds to the balanced sequence

    Example
    --------
    >>> # For a given balanced sequence
    >>> from networkx_algo_common_subtree.utils import graph_str
    >>> open_to_close = {'{': '}', '(': ')', '[': ']'}
    >>> open_to_node = None
    >>> subseq = '({[[]]})[[][]]{{}}'
    >>> # We can convert it into an ordered directed tree
    >>> subtree = seq_to_tree(subseq, open_to_close, open_to_node)
    >>> print(graph_str(subtree))
    ╟── (
    ╎   └─╼ {
    ╎       └─╼ [
    ╎           └─╼ [
    ╟── [
    ╎   ├─╼ [
    ╎   └─╼ [
    ╙── {
        └─╼ {
    """
    nextnode = 0  # only used if open_to_node is not specified
    subtree = OrderedDiGraph()
    stack = []
    for token in subseq:
        if token in open_to_close:
            if open_to_node is None:
                node = nextnode
                nextnode += 1
            else:
                node = open_to_node[token]
            if stack:
                parent_tok, parent_node = stack[-1]
                subtree.add_edge(parent_node, node)
            else:
                subtree.add_node(node)
            if open_to_node is None:
                subtree.nodes[node]["label"] = token
            stack.append((token, node))
        else:
            if not stack:
                raise balanced_sequence.UnbalancedException
            prev_open, prev_node = stack.pop()
            want_close = open_to_close[prev_open]
            if token != want_close:
                raise balanced_sequence.UnbalancedException
    return subtree



from .balanced_isomorphism import (
    longest_common_balanced_isomorphism,
)
from .tree_embedding import tree_to_seq, seq_to_tree
from ._types import OrderedDiGraph


def maximum_common_ordered_subtree_isomorphism(
    tree1, tree2, node_affinity="auto", impl="auto", item_type="auto"
):
    """
    Finds the maximum common subtree-isomorphism between two ordered trees.

    This function computes the maximum-weight common subtrees S1 and S2 between
    two trees T1 and T2. S1 and S2 are isomorphic to subgraphs of T1 and T2
    with maximal size such that S1 and S2 are also isomorphic to each other.

    This function is similar to :func:`maximum_common_ordered_subtree_embedding`
    with the main difference being that returned solution from this function
    will be proper subgraphs (i.e. all edges in the subgraphs will exist in the
    original graph), whereas in the subtree embedding problem the returned
    solutions are allowed to be minors of the input graphs (i.e. edges are
    allowed to be contracted).

    Parameters
    ----------
    tree1, tree2 : OrderedDiGraph
        Trees to find the maximum subtree isomorphism between

    node_affinity : None | str | callable
        Function for to determine if two nodes can be matched. The return is
        interpreted as a weight that is used to break ties. If None then any
        node can match any other node and only the topology is important.
        The default is "eq", which is the same as ``operator.eq``.

    impl : str
        Determines the backend implementation. Defaults to "auto".
        See :func:`balanced_sequence.longest_common_balanced_sequence`
        for details. Other valid options are "iter", "recurse", and
        "iter-cython".

    item_type : str
        Determines the backend data structure used to encode the tree as a
        balanced sequence. Defaults to "auto", other valid options are "chr"
        and "number".

    Returns
    -------
    S1, S2, value: Tuple[OrderedDiGraph, OrderedDiGraph, int]
        The maximum value common subtree isomorphism for each tree with respect
        to the chosen ``node_affinity`` function. The topology of both graphs
        will always be the same, the only difference is that the node labels in
        the first and second embeddings will correspond to ``tree1`` and
        ``tree2`` respectively. When ``node_affinity='eq'`` then embeddings
        should be identical. The last return value is the "size" of the
        solution with respect to ``node_affinity``.

    See Also
    --------
    `maximum_common_ordered_subtree_embedding`
    """
    import networkx as nx

    # Note: checks that inputs are forests are handled by tree_to_seq
    if not isinstance(tree1, OrderedDiGraph):
        raise nx.NetworkXNotImplemented("only implemented for directed ordered trees")
    if not isinstance(tree1, OrderedDiGraph):
        raise nx.NetworkXNotImplemented("only implemented for directed ordered trees")

    if tree1.number_of_nodes() == 0 or tree2.number_of_nodes() == 0:
        raise nx.NetworkXPointlessConcept

    if item_type == "label":
        # If we do allow label, I think the algorithm will work, but the
        # returned tree embeddings will only be embedding wrt to the label
        # structure.
        raise AssertionError(
            "allowing sequences to be specified by the labels breaks assumptions"
        )

    # Convert the trees to balanced sequences.
    # Each sequence will contain each token at most once, this is an important
    # assumption in subsequent steps.
    seq1, open_to_close, node_to_open = tree_to_seq(
        tree1,
        open_to_close=None,
        node_to_open=None,
        item_type=item_type,
        container_type="auto",
    )
    seq2, open_to_close, node_to_open = tree_to_seq(
        tree2, open_to_close, node_to_open, item_type=item_type, container_type="auto"
    )
    open_to_node = {tok: node for node, tok in node_to_open.items()}

    # Solve the longest common balanced sequence problem
    best, value = longest_common_balanced_isomorphism(
        seq1,
        seq2,
        open_to_close,
        open_to_node=open_to_node,
        node_affinity=node_affinity,
        impl=impl,
    )
    subseq1, subseq2 = best

    # Convert the subsequence back into a tree.
    subtree1 = seq_to_tree(subseq1, open_to_close, open_to_node)
    subtree2 = seq_to_tree(subseq2, open_to_close, open_to_node)
    return subtree1, subtree2, value
